@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2014-2015 -- leonerd@leonerd.org.uk
 
 package Event::Distributor;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 use Carp;
 
@@ -101,8 +101,8 @@ sub new
    my $class = shift;
 
    my $self = bless {
-      events      => {},
-      subscribers => {},
+      events           => {},
+      pre_registration => {},
    }, $class;
 
    return $self;
@@ -112,6 +112,21 @@ sub new
 
 =cut
 
+sub _add_event
+{
+   my $self = shift;
+   my ( $name, $event ) = @_;
+
+   $self->{events}{$name} and
+      croak "Cannot declare an event '$name' a second time";
+
+   $self->{events}{$name} = $event;
+
+   if( my $subs = delete $self->{pre_registration}{$name} ) {
+      $event->subscribe( $_ ) for @$subs;
+   }
+}
+
 =head2 $distributor->declare_signal( $name )
 
 Declares a new "signal" event of the given name.
@@ -123,10 +138,7 @@ sub declare_signal
    my $self = shift;
    my ( $name ) = @_;
 
-   $self->{events}{$name} and
-      croak "Cannot declare an event '$name' a second time";
-
-   $self->{events}{$name} = Event::Distributor::Signal->new;
+   $self->_add_event( $name, Event::Distributor::Signal->new );
 }
 
 =head2 $distributor->declare_signal( $name )
@@ -140,10 +152,7 @@ sub declare_query
    my $self = shift;
    my ( $name ) = @_;
 
-   $self->{events}{$name} and
-      croak "Cannot declare an event '$name' a second time";
-
-   $self->{events}{$name} = Event::Distributor::Query->new;
+   $self->_add_event( $name, Event::Distributor::Query->new );
 }
 
 =head2 $distributor->subscribe_async( $name, $code )
@@ -164,8 +173,12 @@ sub subscribe_async
    my $self = shift;
    my ( $name, $code ) = @_;
 
-   # Don't check if the event exists at this time
-   push @{ $self->{subscribers}{$name} }, $code;
+   if( my $event = $self->{events}{$name} ) {
+      $event->subscribe( $code );
+   }
+   else {
+      push @{ $self->{pre_registration}{$name} }, $code;
+   }
 }
 
 =head2 $distributor->subscribe_sync( $name, $code )
@@ -213,9 +226,7 @@ sub fire_async
    my $event = $self->{events}{$name} or
       croak "Cannot fire an event '$name' when it doesn't exist";
 
-   my $subscriberlist = $self->{subscribers}{$name} || [];
-
-   $event->fire( $self, \@args, $subscriberlist );
+   $event->fire( $self, @args );
 }
 
 =head2 $distributor->fire_sync( $name, @args )
